import logging
import os

from .app import name, version
from .camera import AbstractCamera
from .system import AbstractSystem
from flask import Flask, render_template, send_file, Response
from flask_restful import Api, Resource, abort, request

from flask.testing import FlaskClient


API_PREFIX = '/api'

class ActionPiAPI(object): 

    def __init__(self, camera: AbstractCamera, system: AbstractSystem, host: str, port: int, debug=False):
        self._host = host
        self._port = port
        self._debug = debug

        self._app = Flask(__name__)
        self._api = Api(self._app)

        self._camera = camera
        self._system = system

        #Setup routes
        self._api.add_resource(Start, API_PREFIX + '/start', resource_class_args=(camera,))
        self._api.add_resource(Stop, API_PREFIX + '/stop', resource_class_args=(camera,))
        self._api.add_resource(Status, API_PREFIX + '/status', resource_class_args=(camera, system))
        self._api.add_resource(Set, API_PREFIX + '/set', resource_class_args=(camera,))
        self._api.add_resource(Hotspot, API_PREFIX + '/hotspot', resource_class_args=(system, ))
        self._api.add_resource(Halt, API_PREFIX + '/halt', resource_class_args=(system,))
        self._api.add_resource(Reboot, API_PREFIX + '/reboot', resource_class_args=(system,))
        self._api.add_resource(Mount, API_PREFIX + '/mountrw', resource_class_args=(system,))
        self._api.add_resource(Recordings, API_PREFIX + '/recordings', resource_class_args=(camera,))
        self._api.add_resource(Recording, API_PREFIX + '/recording/<string:filename>', resource_class_args=(camera,))

        #Static route
        self._app.add_url_rule('/', '_index', self._index)
        self._app.add_url_rule('/control', '_control', self._control)

        #Preview
        self._app.add_url_rule('/preview', '_preview', self._preview)
    
    def _index(self):
        return render_template('recordings_list_download.html', app={"name":name, "version":version}, file_list=[file for file in os.listdir(self._camera.get_output_dir()) if file.endswith('.h264')])

    def _control(self):
        return render_template('control_panel.html', app={"name":name, "version":version})        

    def _preview(self):
        frame_buff = self._camera.capture_frame()
        if frame_buff is not None:
            return Response((b'--frame\r\nContent-Type: image/jpeg\r\n\r\n' + frame_buff.read() + b'\r\n'),
                            mimetype='multipart/x-mixed-replace; boundary=frame')
        else:
            return Response('', 409)

    def get_test_client(self) -> FlaskClient:
        """
            Only for test purpose
        """
        return self._app.test_client()

    def get_context(self):
        """
            Only for test purpose
        """
        return self._app.app_context()

    def serve(self):
        logging.info("Serving API (debug: %s)", self._debug)
        self._app.run(host=self._host, port=self._port, debug=self._debug)

    def close(self):
        pass

class Start(Resource):
    def __init__(self, camera: AbstractCamera):
        self._camera = camera

    def get(self):
        self._camera.start_recording()
        return '', 204

class Stop(Resource):
    def __init__(self, camera: AbstractCamera):
        self._camera = camera
    
    def get(self):
        self._camera.stop_recording()
        return '', 204

#TODO add more logic if needed in future
class Set(Resource):
    def __init__(self, camera: AbstractCamera):
        self._camera = camera
    
    def get(self, val: int):
        self._camera.change_framerate(val)

class Status(Resource):
    def __init__(self, camera: AbstractCamera, system: AbstractSystem):
        self._camera = camera
        self._system = system
    
    def get(self):
        return {
            'system': {
                'cpu_temperature': self._system.get_cpu_temp(),
                'cpu_load': self._system.get_cpu_percent(),
                'mem_usage': self._system.get_ram_usage(),
                'disk_usage': self._system.get_disks_usage()
            }, 
            'recording': self._camera.is_recording(),
            'framerate': self._camera.get_framerate()
        }

class Halt(Resource):
    def __init__(self, system: AbstractSystem):
        self._system = system

    def get(self):
        logging.info('Shutdown now')
        self._system.halt_system()

class Reboot(Resource):
    def __init__(self, system: AbstractSystem):
        self._system = system

    def get(self):
        logging.info('Reboot now')
        self._system.reboot_system()            

class Hotspot(Resource):
    def __init__(self, system: AbstractSystem):
        self._system = system

    def get(self):
        enable = request.args.get('enable')

        if enable == 'on':
            logging.info('enabling hotspot')
            if not self._system.enable_hotspot():
                abort(500)
        elif enable == 'off':
            logging.info('disabling hotspot')
            if not self._system.disable_hotspot():
                abort(500)
        else:
            logging.error('enable must be "on" or "off", received: %s', enable)
            return 'enable must be "on" or "off"', 400

        return '', 204

class Mount(Resource):
    def __init__(self, system: AbstractSystem):
        self._system = system

    def get(self):
        logging.info('enabling rw file system')
        self._system.mount_rw()
        return '', 204

class Recordings(Resource):
    def __init__(self, camera: AbstractCamera):
        self._camera = camera

    def get(self):
        logging.debug('listing files in %s', self._camera.get_output_dir())
        return os.listdir(self._camera.get_output_dir())

class Recording(Resource):
    def __init__(self, camera: AbstractCamera):
        self._camera = camera

    def get(self, filename: str):
        relative_path = os.path.join(self._camera.get_output_dir(), filename)
        absolute_path = os.path.abspath(relative_path)
        logging.debug('downloading file %s', absolute_path)
        if os.path.exists(absolute_path):
            return send_file(absolute_path)
        else:
            return {'message': 'no file in path {} exists'.format(absolute_path)}, 404
        

    


import logging

from .camera import AbstractCamera
from .system import AbstractSystem

class AbstractIO():
    def __init__(self, camera: AbstractCamera, systen: AbstractSystem, gpio_number: int):
        self._camera = camera
        self._system = systen
        self._gpio_number = gpio_number
    
    def start_monitoring(self):
        logging.info("Start monitoring GPIO {}".format(self._gpio_number))

    def close(self):
        logging.info("Stopping monitoring GPIO {}".format(self._gpio_number))


import logging

from os import path
from io import BytesIO
from abc import ABC, abstractmethod
from threading import RLock

class AbstractCamera(ABC):
    
    def __init__(self,width: int, heigth: int, fps: int, rotation: int, output_dir: str):
        self._lock = RLock()
        with self._lock:
            self._width = width
            self._heigth = heigth
            self._fps = fps
            self._rotation = rotation
            self._output_dir = output_dir
            self._output_file = path.join(output_dir, 'video.h264')

            self._camera = None

    def get_output_dir(self) -> str:
        return self._output_dir

    def start_recording(self):
        logging.info('Recording %ix%i (%i FPS, rotation: %i) video to %s', self._width, self._heigth, self._fps, self._rotation, self._output_file)
        with self._lock:
            self._start()
            

    def change_framerate(self, fps: int):
        logging.debug('Changing FPS to %i', fps)
        with self._lock:
            self.stop_recording()
            self._fps = fps
            self.start_recording()

    def stop_recording(self):
        logging.info('Stopping recording')
        with self._lock:
            self._stop()

    def is_recording(self) -> bool:
        with self._lock:
            return self._recording()

    def get_framerate(self) -> int:
        with self._lock:
            return self._fps

    @abstractmethod
    def capture_frame(self) -> BytesIO:
        pass

    @abstractmethod
    def set_led_status(self, status: bool):
        pass

    @abstractmethod
    def _start(self):
        pass

    @abstractmethod
    def _stop(self):
        pass

    @abstractmethod
    def _recording(self) -> bool:
        pass


import logging
import subprocess
import psutil
import os

from io import BytesIO
from pathlib import Path
from time import sleep

from actionpi import AbstractIO, AbstractCamera, AbstractSystem

try:
    from gpiozero import Button
except (ImportError, ModuleNotFoundError) as e:
    raise ImportError("No module gpiozero installed")

try:
    from picamera import PiCamera
except (ImportError, ModuleNotFoundError) as e:
    raise ImportError("No module picamera installed")

class RaspberryPiCamera(AbstractCamera):

    def __init__(self,width: int, heigth: int, fps: int, rotation: int, output_dir: str):
        super().__init__(width, heigth, fps, rotation, output_dir)
        self._first_run = True
        self._video_file = None

    def _start(self):
        if self._camera is None:
            self._camera = PiCamera(resolution= (self._width, self._heigth), framerate=self._fps)

            self._camera.rotation = self._rotation
            
            if self._first_run:
                fd = os.open(self._output_file, flags=os.O_RDWR|os.O_CREAT|os.O_TRUNC|os.O_SYNC)
                if fd != -1:
                    self._video_file = open(fd, 'wb', buffering=0)
                    self._first_run = False
                else:
                    raise OSError()
            else:
                fd = os.open(self._output_file, flags=os.O_RDWR|os.O_CREAT|os.O_APPEND|os.O_SYNC)
                if fd != -1:
                    self._video_file = open(fd, 'ab', buffering=0)
                else:
                    raise OSError()

            self._camera.start_recording(self._video_file, format='h264')

    def _stop(self):
        if self._camera is not None:
            self._camera.stop_recording()
            self._camera.close()
            self._video_file.close()
            self._video_file = None
            self._camera = None

    def _recording(self) -> bool:
        return (self._camera is not None) and (self._camera.recording)

    def get_framerate(self) -> int:
        if self._camera is not None:
            return int(self._camera.framerate)
        else:
            return 0
    
    def set_led_status(self, status: bool):
        if self._camera is not None:
            self._camera.led = status

    def capture_frame(self) -> BytesIO:
        if self._camera is not None:
            capture_stream = BytesIO()
            self._camera.capture(capture_stream, 'jpeg', resize=(320, 240), use_video_port=True)
            capture_stream.seek(0)
            return capture_stream

class RaspberryPiSystem(AbstractSystem):
    def get_cpu_temp(self) -> float:
        try:
            temp = int(open('/sys/class/thermal/thermal_zone0/temp').read()) / 1e3
            return temp
        except IOError:
            return 0

    def get_cpu_percent(self) -> int:
        return psutil.cpu_percent(interval=None)

    def get_disks_usage(self) -> list:
        disk_usages = list()
        for part in psutil.disk_partitions(all=False):
            mountpoint = part.mountpoint
            usage = psutil.disk_usage(mountpoint)
            disk_usages.append({
                'mountpoint' : mountpoint,
                'percent' : usage.percent,
                'total' : usage.total,
                'used' : usage.used
            })
        return disk_usages
    
    def get_ram_usage(self) -> int:
        return psutil.virtual_memory().percent

    def halt_system(self):
        subprocess.run(["shutdown", "-H", "now"])

    def reboot_system(self):
        subprocess.run(["shutdown", "-r", "now"])

    def enable_hotspot(self) -> bool:
        Path('/boot/wifi_hotspot').touch()
        try:
            Path('/boot/wifi_client').unlink()
        except FileNotFoundError:
            pass
        return True

    def disable_hotspot(self) -> bool:
        Path('/boot/wifi_client').touch()
        try:
            Path('/boot/wifi_hotspot').unlink()
        except FileNotFoundError:
            pass
        return True

    def get_hw_revision(self) -> str:
        hw_rev = "00000"
        try:
            f = open('/proc/cpuinfo','r')
            for line in f:
                if line[0:8]=='Revision':
                    hw_rev = line[10:]
            f.close()
        except:
            hw_rev = "00000"

        return hw_rev

    def get_serial(self) -> str:
        # Extract serial from cpuinfo file
        cpuserial = "0000000000000000"
        try:
            f = open('/proc/cpuinfo','r')
            for line in f:
                if line[0:6]=='Serial':
                    cpuserial = line[10:26]
            f.close()
        except:
            cpuserial = "ERROR000000000"

        return cpuserial

    def mount_rw(self):
        Path('/boot/rw').touch()

class RaspberryPiIO(AbstractIO):

    def __init__(self, camera: AbstractCamera, system: AbstractSystem, gpio_number: int):
        super().__init__(camera, system, gpio_number)
        self.button = Button(gpio_number, bounce_time=0.2)
    
    def start_monitoring(self):
        super().start_monitoring()
        self.button.when_pressed = self._start_recording_handler
        self.button.when_released = self._stop_recording_handler
        
        if self.button.is_pressed:
            self._camera.start_recording()

    def _start_recording_handler(self):
        self._camera.start_recording()

    def _stop_recording_handler(self):
        self._camera.stop_recording()
        sleep(1)
        self._system.halt_system()

    def close(self):
        super().close()
        self.button.close()
        self.button = None


from . import AbstractCamera, AbstractIO, AbstractSystem

class ActionPiFactory(object):

    @staticmethod   
    def get_system(board: str) -> AbstractSystem:
        if board == 'raspberrypi':
            from .imp.raspberrypi import RaspberryPiSystem
            return RaspberryPiSystem()
        else:
            raise RuntimeError('board system not supported:' + board)

    @staticmethod
    def get_io(board: str, camera: AbstractCamera, system: AbstractSystem, gpio: int) -> AbstractIO:
        if board == 'raspberrypi':
            from .imp.raspberrypi import RaspberryPiIO
            return RaspberryPiIO(camera, system, gpio)
        else:
            raise RuntimeError('board system not supported:' + board)

    @staticmethod
    def get_camera(board: str, width: int, heigth:int, fps:int, rotation: int, output_dir: str) -> AbstractCamera:
        if board == 'raspberrypi':
            from .imp.raspberrypi import RaspberryPiCamera
            return RaspberryPiCamera(width, heigth, fps, rotation, output_dir)
        else:
            raise RuntimeError('board system not supported:' + board)